from typing import Any, Callable, List, Optional, Union

from discord.reaction import Reaction

from discord import Embed, Member, Message, User

from .prompt import Prompt

__all__ = (
    "BaseMixin",
    "ResendMixin",
    "DeleteUserResponseMixin",
    "ClearReactionsMixin",
)


class BaseMixin:
    message: Message
    embed: Embed
    user: Union[User, Member]
    prompts: List[Prompt]
    yes_emoji: str
    no_emoji: str

    update_embed: Callable[[Prompt, Optional[Prompt], Any], Any]
    handle_message_input: Callable[[Message, bool], Any]
    handle_reaction_input: Callable[[Reaction], Any]


class ResendMixin(BaseMixin):
    async def update_message(
        self, prompt: Prompt, next_prompt: Optional[Prompt], result
    ) -> None:
        self.update_embed(prompt, next_prompt, result)
        self._message = await self.message.channel.send(embed=self.embed)


class DeleteUserResponseMixin(BaseMixin):
    async def handle_message_input(self, message: Message, input_error: bool):
        await message.delete()


class ClearReactionsMixin(BaseMixin):
    async def handle_reaction_input(self, reaction: Reaction):
        await reaction.message.clear_reactions()



from dataclasses import dataclass
from typing import Callable, Optional

__all__ = ("Prompt",)


@dataclass
class Prompt:
    title: str
    max_retries: int = 5
    key: Optional[str] = None
    description: Optional[str] = None
    check: Optional[Callable] = None
    post_check: Optional[Callable] = None
    timeout: Optional[int] = None
    reaction_interface: Optional[bool] = None
    res_type: type = str

    def get_waiting_message(self):
        return f"{self.description}\nWaiting...." if self.description else "Waiting..."



from collections import namedtuple

_VersionInfo = namedtuple("_VersionInfo", "major minor patch")
version_info = _VersionInfo(0, 1, 1)

__version__ = ".".join(str(v) for v in version_info)
__license__ = "MIT"
__author__ = "FalseDev"
__title__ = "discord-ext-wizard"
__copyright__ = "Copyright 2021 {}".format(__author__)
__uri__ = "https://github.com/{}/{}".format(__author__, __title__)

from .checks import *
from .mixins import *
from .prompt import *
from .wizard import *



import asyncio
import inspect
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from discord.abc import Messageable
from discord.ext.commands import AutoShardedBot, Bot, CommandError, UserInputError

from discord import Color, Embed, Emoji, Member, Message, Reaction, User

from .converters import ConverterMapping
from .prompt import Prompt

__all__ = ("WizardFailure", "EmbedWizard")

BotType = Union[Bot, AutoShardedBot]


async def maybe_async(func, *args, **kwargs):
    res = func(*args, **kwargs)
    if inspect.iscoroutine(res):
        return await res
    return res


class WizardFailure(CommandError):
    pass


class EmbedWizard:
    def __init__(
        self,
        *,
        title: str = "Wizard",
        prompts: List[Prompt],
        bot: BotType,
        channel: Messageable,
        user: Union[User, Member],
        converters: Dict[type, Callable] = None,
        default_timeout=60,
        yes_emoji: str = "\u2705",
        no_emoji: str = "\u274c",
    ) -> None:
        self.title: str = title
        self.prompts: List[Prompt] = prompts
        self.channel: Messageable = channel
        self.user: Union[User, Member] = user
        self.bot: BotType = bot
        self._message: Optional[Message] = None
        self.converters: ConverterMapping = ConverterMapping(
            {} if converters is None else converters
        )
        self.default_timeout: int = default_timeout
        self.results: List[Any] = []

        # Emojis
        self.yes_emoji: str = yes_emoji
        self.no_emoji: str = no_emoji

    # Properties

    @property
    def message(self) -> Message:
        if not self._message:
            raise RuntimeError("Wizard hasn't started")
        return self._message

    @property
    def embed(self) -> Embed:
        return self.message.embeds[0]

    # Default checks

    def default_message_check(self, m: Message) -> bool:
        return m.channel == self.channel and self.user == m.author

    def default_reaction_check(self, r: Reaction, u: User) -> bool:
        return u == self.user and self.message == r.message

    # Check combiners

    def combine_message_checks(self, check: Optional[Callable]):
        if not check:
            return self.default_message_check

        def combined_check(m: Message):
            return self.default_message_check(m) and check(m)

        return combined_check

    def combine_reaction_checks(self, check: Optional[Callable]):
        if not check:
            return self.default_reaction_check

        def combined_check(r: Reaction, u: User):
            return self.default_reaction_check(r, u) and check(r, u)

        return combined_check

    # Helpers

    def get_timeout(self, prompt_timeout):
        return prompt_timeout or self.default_timeout

    # Input methods

    async def get_actual_input(self, prompt: Prompt):
        if not prompt.reaction_interface and not issubclass(
            prompt.res_type, (Reaction, Emoji)
        ):
            return await self.get_message_input(prompt)
        return await self.get_reaction_input(prompt)

    async def get_message_input(self, prompt: Prompt):
        converter = self.converters[prompt.res_type]
        check = self.combine_message_checks(prompt.check)
        timeout = self.get_timeout(prompt.timeout)
        msg = await self.bot.wait_for("message", check=check, timeout=timeout)
        ctx = await self.bot.get_context(msg)

        input_error: bool = True
        try:
            result = await maybe_async(converter, ctx, msg.content)
            input_error = False
            return result
        except:
            raise
        finally:
            await self.handle_message_input(msg, input_error)

    async def get_reaction_input(self, prompt: Prompt):
        reaction = await self.get_actual_reaction_input(prompt)
        await self.handle_reaction_input(reaction[0])
        return reaction[1]

    async def get_actual_reaction_input(self, prompt: Prompt) -> Tuple[Reaction, Any]:
        timeout = self.get_timeout(prompt.timeout)
        if prompt.res_type in (Reaction, Emoji):
            check = self.combine_reaction_checks(prompt.check)
            reaction, _ = await self.bot.wait_for(
                "reaction_add", check=check, timeout=timeout
            )
            if prompt.res_type is Reaction:
                return reaction, reaction
            return reaction, reaction.emoji

        if prompt.res_type is bool:
            await self.message.add_reaction(self.yes_emoji)
            await self.message.add_reaction(self.no_emoji)

            def bool_check(reaction: Reaction, user: User):
                return reaction.emoji in (self.yes_emoji, self.no_emoji)

            check = self.combine_reaction_checks(bool_check)
            reaction, _ = await self.bot.wait_for(
                "reaction_add", check=check, timeout=timeout
            )
            return reaction, reaction.emoji == self.yes_emoji

        raise NotImplementedError(
            "Reaction input for given res_type {0.res_type} is unavailable!".format(
                prompt
            )
        )

    # Front Facing methods
    async def get_input(self, prompt: Prompt):
        retries = 0
        while retries < prompt.max_retries:
            try:
                result = await self.get_actual_input(prompt)
                if prompt.post_check:
                    await maybe_async(prompt.post_check, self, result)
                return result
            except UserInputError as e:
                await self.handle_error(e)
            retries += 1
        raise WizardFailure("Maximum retries reached")

    async def start(self) -> None:
        if self._message:
            raise RuntimeError("Wizard already started")
        embed = Embed(title=self.title, color=Color.green())
        prompt = self.prompts[0]
        embed.add_field(
            name=prompt.title,
            value=prompt.get_waiting_message(),
            inline=False,
        )
        self._message = await self.channel.send(embed=embed)

    async def update_message(
        self, prompt: Prompt, next_prompt: Optional[Prompt], result
    ) -> None:
        self.update_embed(prompt, next_prompt, result)
        await self.message.edit(embed=self.embed)

    def update_embed(
        self, prompt: Prompt, next_prompt: Optional[Prompt], result
    ) -> None:
        embed = self.embed
        embed.remove_field(-1)
        embed.add_field(name=prompt.title, value=self.to_str(result), inline=False)
        if next_prompt:
            embed.add_field(
                name=next_prompt.title,
                value=next_prompt.get_waiting_message(),
                inline=False,
            )
        else:
            embed.add_field(name="End", value="Wizard has ended", inline=False)

    async def run(self):
        await self.start()
        for index in range(len(self.prompts)):
            prompt = self.prompts[index]
            try:
                result = await self.get_input(prompt)
            except asyncio.TimeoutError as e:
                return await self.on_timeout(prompt, e)
            self.results.append(result)
            next_prompt = (
                self.prompts[index + 1] if len(self.prompts) > (index + 1) else None
            )
            await self.update_message(prompt, next_prompt, result)
        return self.results

    async def handle_error(self, e: Exception):
        await self.channel.send(e)

    async def on_timeout(self, prompt: Prompt, e: asyncio.TimeoutError):
        await self.handle_error(
            WizardFailure("The wizard has failed due to no response for a long time.")
        )
        raise e

    async def handle_message_input(self, message: Message, input_error: bool):
        pass

    async def handle_reaction_input(self, reaction: Reaction):
        pass

    def to_str(self, result):
        if hasattr(result, "mention"):
            return result.mention
        if isinstance(result, Message):
            return "Message with ID {0.id} in {0.channel.mention}".format(result)
        return str(result)



from typing import Optional, Union

from discord.ext.commands import UserInputError

from discord import Emoji, Member, PartialEmoji, Reaction, Role

from .wizard import EmbedWizard

__all__ = (
    "IntRangeCheck",
    "StringLengthCheck",
    "CurrentGuildEmojiCheck",
    "CurrentGuildReactionCheck",
    "RoleBelowMemberCheck",
)


class IntRangeCheck:
    def __init__(
        self,
        min_value: int = None,
        max_value: int = None,
        *,
        too_large_message: str = "Value too small, must be more than {}",
        too_small_message: str = "Value too large, must be less than {}",
    ) -> None:
        if min_value is None and max_value is None:
            raise RuntimeError("Both min and max can't be None")

        self.min_value: Optional[int] = min_value
        self.max_value: Optional[int] = max_value
        self.too_large_message: str = too_large_message
        self.too_small_message: str = too_small_message

    def __call__(self, wizard: EmbedWizard, value: int) -> None:
        if self.min_value is not None and self.min_value > value:
            raise UserInputError(self.too_small_message.format(self.min_value))

        if self.max_value is not None and self.max_value < value:
            raise UserInputError(self.too_large_message.format(self.max_value))


class StringLengthCheck(IntRangeCheck):
    def __init__(
        self,
        min_len: int = None,
        max_len: int = None,
        *,
        too_small_message: str = "Value too short, must be more than {}",
        too_large_message: str = "Value too large, must be less than {}",
    ) -> None:
        super().__init__(
            min_value=min_len,
            max_value=max_len,
            too_large_message=too_large_message,
            too_small_message=too_small_message,
        )

    def __call__(self, wizard: EmbedWizard, value: str) -> None:
        super().__call__(wizard, len(value))


class CurrentGuildEmojiCheck:
    def __init__(self, *, error_message: str = "The emoji is from a different server!"):
        self.error_message: str = error_message

    def __call__(
        self,
        wizard: EmbedWizard,
        emoji: Union[Emoji, PartialEmoji, str],
    ) -> None:
        if isinstance(emoji, str):
            return

        if isinstance(emoji, PartialEmoji):
            if emoji.is_unicode_emoji():
                return
            raise UserInputError(self.error_message)

        if isinstance(wizard.user, Member) and emoji.guild_id == wizard.user.guild.id:
            return
        raise UserInputError(self.error_message)


class CurrentGuildReactionCheck(CurrentGuildEmojiCheck):
    def __init__(
        self,
        *,
        error_message: str = "The emoji used in the reaction is from a different server!",
    ):
        self.error_message: str = error_message

    def __call__(self, wizard: EmbedWizard, reaction: Reaction) -> None:
        super().__call__(wizard, reaction.emoji)


class RoleBelowMemberCheck:
    def __init__(
        self, *, error_message: str = "The role selected is above your highest role!"
    ) -> None:
        self.error_message: str = error_message

    def __call__(self, wizard: EmbedWizard, role: Role) -> None:
        user = wizard.user
        if not isinstance(user, Member):
            raise UserInputError(self.error_message)

        if user.guild.owner == user:
            return

        if role.position >= user.top_role.position:
            raise UserInputError(self.error_message)



from typing import Awaitable, Callable, Dict, Mapping

from discord.ext.commands import BadArgument, BadBoolArgument, Context
from discord.ext.commands import converter as discord_converters


async def convert_int(_, content: str) -> int:
    try:
        return int(content)
    except ValueError as exc:
        raise BadArgument(f"Invalid int value: {content}") from exc


async def convert_bool(_, content: str) -> bool:
    s = content.lower()
    if s in ("yes", "y", "1", "true", "t"):
        return True
    if s in ("no", "n", "0", "false", "f"):
        return False
    raise BadBoolArgument(content)


async def convert_str(_, content: str) -> str:
    return content


class ConverterMapping:
    preset_converters: Dict[type, Callable] = {
        bool: convert_bool,
        str: convert_str,
        int: convert_int,
    }

    def __init__(self, converters: Mapping[type, Callable]) -> None:
        self.converters = converters

    def __getitem__(self, key: type) -> Callable[[Context, str], Awaitable]:
        if key in self.converters:
            return self.converters[key]

        if key in self.preset_converters:
            return self.preset_converters[key]

        try:
            return getattr(discord_converters, key.__name__ + "Converter")().convert
        except AttributeError as exc:
            raise RuntimeError("Couldn't find valid converter") from exc



from setuptools import setup

version = "0.1.1"

with open("README.md") as f:
    long_description = f.read()

requirements = ["discord.py>=1.5.0"]

setup(
    name="discord-ext-wizard",
    author="FalseDev",
    author_email="59121948+FalseDev@users.noreply.github.com",
    version=version,
    url="https://github.com/FalseDev/discord-ext-wizard",
    packages=["discord.ext.wizard"],
    license="MIT",
    description="A Discord.py extension that makes setup wizards too easy.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    install_requires=requirements,
    extras_require=None,
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
    ],
)
